###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from pyomo.environ import (
    Var,
    Suffix,
    NonNegativeReals,
    Reals,
    units as pyunits,
    check_optimal_termination,
)
from pyomo.common.config import Bool, ConfigBlock, ConfigValue, In
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers import get_solver
from idaes.core.util.constants import Constants
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from watertap.core import ControlVolume0DBlock, InitializationMixin
from watertap.costing.unit_models.electrolyzer import cost_electrolyzer

__author__ = "Hunter Barber"

_log = idaeslog.getLogger(__name__)


# ---------------------------------------------------------------------
@declare_process_block_class("Electrolyzer")
class ElectrolyzerData(InitializationMixin, UnitModelBlockData):
    """
    Faradaic conversion electrolyzer model
    developed specifically for chlor alkali electrolysis using a membrane electrolyzer
    this model development is ongoing
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The electrolyzer unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The electrolyzer unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "is_isothermal",
        ConfigValue(
            default=True,
            domain=Bool,
            description="""Assume isothermal conditions for control volume(s); energy_balance_type must be EnergyBalanceType.none,
    **default** - True.""",
        ),
    )

    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.none,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.none.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    def _validate_config(self):
        if (
            self.config.is_isothermal
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            raise ConfigurationError(
                "If the isothermal assumption is used then the energy balance type must be none"
            )

    # ---------------------------------------------------------------------

    def build(self):
        super().build()

        # create blank scaling factors to be populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)
        # get default units from property package
        units_meta = self.config.property_package.get_metadata().get_derived_units
        # Check configs for errors
        self._validate_config()

        # create methods for use in custom_molar_term of control volumes
        self.custom_reaction_anode = Var(
            self.flowsheet().time,
            self.config.property_package.component_list,
            initialize=1,
            bounds=(None, None),
            domain=Reals,
            units=units_meta("amount") * units_meta("time") ** -1,
        )

        def custom_method_anode(t, j):
            return self.custom_reaction_anode[t, j]

        self.custom_reaction_cathode = Var(
            self.flowsheet().time,
            self.config.property_package.component_list,
            initialize=1,
            bounds=(None, None),
            domain=Reals,
            units=units_meta("amount") * units_meta("time") ** -1,
        )

        def custom_method_cathode(t, j):
            return self.custom_reaction_cathode[t, j]

        # build control volume
        self.anolyte = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )
        self.catholyte = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )
        for control_volume in [self.anolyte, self.catholyte]:
            control_volume.add_state_blocks(has_phase_equilibrium=False)
            control_volume.add_energy_balances(
                balance_type=self.config.energy_balance_type,
                has_enthalpy_transfer=False,
            )
            if self.config.is_isothermal:
                control_volume.add_isothermal_assumption()
                control_volume.add_momentum_balances(
                    balance_type=self.config.momentum_balance_type,
                    has_pressure_change=False,
                )
        self.anolyte.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=True,
            custom_molar_term=custom_method_anode,
        )
        self.catholyte.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=True,
            custom_molar_term=custom_method_cathode,
        )

        # Add ports
        self.add_inlet_port(name="anolyte_inlet", block=self.anolyte)
        self.add_outlet_port(name="anolyte_outlet", block=self.anolyte)
        self.add_inlet_port(name="catholyte_inlet", block=self.catholyte)
        self.add_outlet_port(name="catholyte_outlet", block=self.catholyte)

        # ---------------------------------------------------------------------
        # membrane properties TODO: move to generic flat membrane block

        self.membrane_area = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") ** 2,
            doc="membrane area",
        )
        self.membrane_current_density = Var(
            initialize=1e4,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("current") * units_meta("length") ** -2,
            doc="membrane current density",
        )
        self.membrane_ion_transport_number = Var(
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=0,
            domain=Reals,
            units=pyunits.dimensionless,
            doc="ion transport number of species passing from the anode to the cathode"
            " through the membrane normalized to 1 electron",
        )

        # ---------------------------------------------------------------------
        # anode properties TODO: move to generic anode block

        self.anode_area = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") ** 2,
            doc="anode area",
        )
        self.anode_current_density = Var(
            initialize=1e4,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("current") * units_meta("length") ** -2,
            doc="anode current density",
        )
        self.anode_electrochem_potential = Var(
            initialize=1,
            domain=Reals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3
            * units_meta("current") ** -1,
            doc="electrochemical potential of the reaction at the anode",
        )
        self.anode_overpotential = Var(
            initialize=0.1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3
            * units_meta("current") ** -1,
            doc="anode overpotential",
        )
        self.anode_stoich = Var(
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=0,
            domain=Reals,
            units=pyunits.dimensionless,
            doc="stoichiometry of the reaction at the anode normalized to 1 electron",
        )

        # ---------------------------------------------------------------------
        # cathode properties TODO: move to generic cathode block

        self.cathode_area = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") ** 2,
            doc="anode area",
        )
        self.cathode_current_density = Var(
            initialize=1e4,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("current") * units_meta("length") ** -2,
            doc="anode current density",
        )
        self.cathode_electrochem_potential = Var(
            initialize=1,
            domain=Reals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3
            * units_meta("current") ** -1,
            doc="electrochemical potential of the reaction at the cathode",
        )
        self.cathode_overpotential = Var(
            initialize=0.1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3
            * units_meta("current") ** -1,
            doc="cathode overpotential",
        )
        self.cathode_stoich = Var(
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=0,
            domain=Reals,
            units=pyunits.dimensionless,
            doc="stoichiometry of the reaction at the cathode normalized to 1 electron",
        )

        # ---------------------------------------------------------------------
        # electrolyzer cell design

        self.current = Var(
            initialize=1e5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("current"),
            doc="DC current supplied to the cell",
        )
        self.voltage_cell = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3
            * units_meta("current") ** -1,
            doc="applied voltage to the cell",
        )
        self.resistance = Var(
            initialize=1e-5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3
            * units_meta("current") ** -2,
            doc="ohmic resistance, the sum of membrane and electrolyte resistances",
        )
        self.power = Var(
            initialize=1e5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("power"),
            doc="power",
        )

        # ---------------------------------------------------------------------
        # performance variables

        self.voltage_reversible = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3
            * units_meta("current") ** -1,
            doc="reversible voltage",
        )
        self.electron_flow = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("amount") * units_meta("time") ** -1,
            doc="electrons contributing to electrolysis reactions",
        )
        self.efficiency_current = Var(
            initialize=0.9,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="current efficiency",
        )
        self.efficiency_voltage = Var(
            initialize=0.9,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="voltage efficiency",
        )
        self.efficiency_power = Var(
            initialize=0.81,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="power efficiency",
        )

        # ---------------------------------------------------------------------
        # relieve exhaustive fixing of indexed variables

        # fix stoichiometry to zero, to be overwritten when constructing flowsheet
        for p in self.config.property_package.phase_list:
            for j in self.config.property_package.component_list:
                self.anode_stoich[p, j].fix(0)
                self.cathode_stoich[p, j].fix(0)
                self.membrane_ion_transport_number[p, j].fix(0)

        # ---------------------------------------------------------------------
        # membrane calculations

        @self.Constraint(doc="membrane current density")
        def eq_membrane_current_density(b):
            return b.membrane_current_density * b.membrane_area == b.current

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="ion permeation through the membrane",
        )
        def eq_membrane_permeation(b, t, p, j):
            return (
                b.anolyte.mass_transfer_term[t, p, j]
                == -b.electron_flow * b.membrane_ion_transport_number[p, j]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="equating membrane permeation between the two control volumes",
        )
        def eq_mass_transfer_membrane(b, t, p, j):
            return (
                b.catholyte.mass_transfer_term[t, p, j]
                == -b.anolyte.mass_transfer_term[t, p, j]
            )

        # ---------------------------------------------------------------------
        # anode calculations

        @self.Constraint(doc="anode current density")
        def eq_anode_current_density(b):
            return b.anode_current_density * b.anode_area == b.current

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="molar generation of species according the anode electrolysis reaction",
        )
        def eq_custom_reaction_anode(b, t, p, j):
            return (
                b.custom_reaction_anode[t, j] == b.electron_flow * b.anode_stoich[p, j]
            )

        # ---------------------------------------------------------------------
        # cathode calculations

        @self.Constraint(doc="cathode current density")
        def eq_cathode_current_density(b):
            return b.cathode_current_density * b.cathode_area == b.current

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="molar generation of species according the cathode electrolysis reaction",
        )
        def eq_custom_reaction_cathode(b, t, p, j):
            return (
                b.custom_reaction_cathode[t, j]
                == b.electron_flow * b.cathode_stoich[p, j]
            )

        # ---------------------------------------------------------------------
        # performance caclculations

        @self.Constraint(
            doc="reversible voltage based on the difference in electrochemical potentials"
        )
        def eq_voltage_reversible(b):
            return (
                b.voltage_reversible
                == b.anode_electrochem_potential - b.cathode_electrochem_potential
            )

        @self.Constraint(
            doc="cell voltage as the sum of overpotentials and ohms law relation"
        )
        def eq_voltage_cell(b):
            return b.voltage_cell == (
                b.voltage_reversible
                + b.anode_overpotential
                + b.cathode_overpotential
                + b.current * b.resistance
            )

        @self.Constraint(doc="energy consumption")
        def eq_power(b):
            return b.power == b.voltage_cell * b.current

        @self.Constraint(doc="electrons contributing to reactions")
        def eq_electron_flow(b):
            return (
                b.electron_flow
                * pyunits.convert(
                    Constants.faraday_constant,
                    to_units=units_meta("time")
                    * units_meta("current")
                    / units_meta("amount"),
                )
                == b.current * b.efficiency_current
            )

        @self.Constraint(doc="voltage efficiency")
        def eq_efficiency_voltage(b):
            return b.voltage_reversible == b.voltage_cell * b.efficiency_voltage

        @self.Constraint(doc="energy efficiency")
        def eq_efficiency_power(b):
            return b.efficiency_power == b.efficiency_current * b.efficiency_voltage

    # ---------------------------------------------------------------------
    # initialize method

    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # set state_args from inlet state

        for control_volume in [self.anolyte, self.catholyte]:
            if state_args is None:
                state_args = {}
                state_dict = control_volume.properties_in[
                    self.flowsheet().config.time.first()
                ].define_port_members()

                for k in state_dict.keys():
                    if state_dict[k].is_indexed():
                        state_args[k] = {}
                        for m in state_dict[k].keys():
                            state_args[k][m] = state_dict[k][m].value
                    else:
                        state_args[k] = state_dict[k].value

        # initialize anolyte control volume
        flags_anolyte = self.anolyte.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )

        # initialize catholyte control volume
        flags_catholyte = self.catholyte.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )

        init_log.info_high("Initialization Step 1 Complete.")
        # --------------------------------------------------------------------
        # solve unit

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))
        # ---------------------------------------------------------------------
        # release inlet state

        self.anolyte.release_state(flags_anolyte, outlvl + 1)
        self.catholyte.release_state(flags_catholyte, outlvl + 1)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    # ---------------------------------------------------------------------
    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        anolyte_in = self.anolyte.properties_in[time_point]
        anolyte_out = self.anolyte.properties_out[time_point]
        catholyte_in = self.catholyte.properties_in[time_point]
        catholyte_out = self.catholyte.properties_out[time_point]

        # unit model variables
        var_dict["membrane area"] = self.membrane_area
        var_dict["membrane current density"] = self.membrane_current_density
        var_dict["anode area"] = self.membrane_area
        var_dict["anode current density"] = self.membrane_current_density
        var_dict["cathode area"] = self.membrane_area
        var_dict["cathode current density"] = self.membrane_current_density
        var_dict["DC current supplied"] = self.current
        var_dict["applied voltage"] = self.voltage_cell
        var_dict["ohmic resistance"] = self.resistance
        var_dict["power"] = self.power
        var_dict["reversible voltage"] = self.voltage_reversible
        var_dict["electrons passed"] = self.electron_flow
        var_dict["current efficiency"] = self.efficiency_current
        var_dict["voltage efficiency"] = self.efficiency_voltage
        var_dict["power efficiency"] = self.efficiency_power

        # loop through desired state block properties indexed by [phase, comp]
        phase_comp_prop_dict = {
            "flow_mol_phase_comp": "molar flow rate",
            "conc_mass_phase_comp": "mass concentration",
        }
        for prop_name, prop_str in phase_comp_prop_dict.items():
            for j in self.config.property_package.component_list:
                if anolyte_in.is_property_constructed(prop_name):
                    var_dict[f"{prop_str} of {j} @ anolyte inlet"] = getattr(
                        anolyte_in, prop_name
                    )["Liq", j]
                if anolyte_out.is_property_constructed(prop_name):
                    var_dict[f"{prop_str} of {j} @ anolyte outlet"] = getattr(
                        anolyte_out, prop_name
                    )["Liq", j]
                if catholyte_in.is_property_constructed(prop_name):
                    var_dict[f"{prop_str} of {j} @ catholyte inlet"] = getattr(
                        catholyte_in, prop_name
                    )["Liq", j]
                if catholyte_out.is_property_constructed(prop_name):
                    var_dict[f"{prop_str} of {j} @ catholyte outlet"] = getattr(
                        catholyte_out, prop_name
                    )["Liq", j]

        # loop through desired state block properties indexed by [phase]
        phase_prop_dict = {
            "flow_vol_phase": "volumetric flow rate",
        }
        for prop_name, prop_str in phase_prop_dict.items():
            if anolyte_in.is_property_constructed(prop_name):
                var_dict[f"{prop_str} @ anolyte inlet"] = getattr(
                    anolyte_in, prop_name
                )["Liq"]
            if anolyte_out.is_property_constructed(prop_name):
                var_dict[f"{prop_str} @ anolyte outlet"] = getattr(
                    anolyte_out, prop_name
                )["Liq"]
            if catholyte_in.is_property_constructed(prop_name):
                var_dict[f"{prop_str} @ catholyte inlet"] = getattr(
                    catholyte_in, prop_name
                )["Liq"]
            if catholyte_out.is_property_constructed(prop_name):
                var_dict[f"{prop_str} @ catholyte outlet"] = getattr(
                    catholyte_out, prop_name
                )["Liq"]

        return {"vars": var_dict}

    # ---------------------------------------------------------------------
    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Anolyte Inlet": self.anolyte_inlet,
                "Anolyte Outlet": self.anolyte_outlet,
                "Catholyte Inlet": self.catholyte_inlet,
                "Catholyte Outlet": self.catholyte_outlet,
            },
            time_point=time_point,
        )

    # ---------------------------------------------------------------------
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # ---------------------------------------------------------------------
        # fixed sf

        if iscale.get_scaling_factor(self.membrane_current_density) is None:
            iscale.set_scaling_factor(self.membrane_current_density, 1e-4)

        if iscale.get_scaling_factor(self.anode_current_density) is None:
            iscale.set_scaling_factor(self.anode_current_density, 1e-4)

        if iscale.get_scaling_factor(self.anode_electrochem_potential) is None:
            iscale.set_scaling_factor(self.anode_electrochem_potential, 1)

        if iscale.get_scaling_factor(self.anode_overpotential) is None:
            iscale.set_scaling_factor(self.anode_overpotential, 1e1)

        if iscale.get_scaling_factor(self.cathode_current_density) is None:
            iscale.set_scaling_factor(self.cathode_current_density, 1e-4)

        if iscale.get_scaling_factor(self.cathode_electrochem_potential) is None:
            iscale.set_scaling_factor(self.cathode_electrochem_potential, 1)

        if iscale.get_scaling_factor(self.cathode_overpotential) is None:
            iscale.set_scaling_factor(self.cathode_overpotential, 1e1)

        if iscale.get_scaling_factor(self.voltage_cell) is None:
            iscale.set_scaling_factor(self.voltage_cell, 1)

        if iscale.get_scaling_factor(self.resistance) is None:
            iscale.set_scaling_factor(self.resistance, 1e5)

        if iscale.get_scaling_factor(self.voltage_reversible) is None:
            iscale.set_scaling_factor(self.voltage_reversible, 1)

        if iscale.get_scaling_factor(self.efficiency_current) is None:
            iscale.set_scaling_factor(self.efficiency_current, 1)

        if iscale.get_scaling_factor(self.efficiency_voltage) is None:
            iscale.set_scaling_factor(self.efficiency_voltage, 1)

        if iscale.get_scaling_factor(self.efficiency_power) is None:
            iscale.set_scaling_factor(self.efficiency_power, 1)

        for p in self.config.property_package.phase_list:
            for j in self.config.property_package.component_list:
                iscale.set_scaling_factor(self.membrane_ion_transport_number[p, j], 1)
                iscale.set_scaling_factor(self.anode_stoich[p, j], 1)
                iscale.set_scaling_factor(self.cathode_stoich[p, j], 1)

        for t in self.flowsheet().time:
            for j in self.config.property_package.component_list:
                if (
                    iscale.get_scaling_factor(self.custom_reaction_cathode[t, j])
                    is None
                ):
                    iscale.set_scaling_factor(self.custom_reaction_cathode[t, j], 1)

                if iscale.get_scaling_factor(self.custom_reaction_anode[t, j]) is None:
                    iscale.set_scaling_factor(self.custom_reaction_anode[t, j], 1)

        # ---------------------------------------------------------------------
        # calculated sf

        # determine sf as a function of current, therefore a user should
        # first scale current if the model is poorly scaled
        if iscale.get_scaling_factor(self.current) is None:
            sf_current = 1e-5
            iscale.set_scaling_factor(self.current, sf_current)
        else:
            sf_current = iscale.get_scaling_factor(self.current)

        if iscale.get_scaling_factor(self.membrane_area) is None:
            iscale.set_scaling_factor(
                self.membrane_area,
                sf_current
                * iscale.get_scaling_factor(self.membrane_current_density) ** -1,
            )

        if iscale.get_scaling_factor(self.anode_area) is None:
            iscale.set_scaling_factor(
                self.anode_area,
                sf_current
                * iscale.get_scaling_factor(self.anode_current_density) ** -1,
            )

        if iscale.get_scaling_factor(self.cathode_area) is None:
            iscale.set_scaling_factor(
                self.cathode_area,
                sf_current
                * iscale.get_scaling_factor(self.cathode_current_density) ** -1,
            )

        if iscale.get_scaling_factor(self.power) is None:
            iscale.set_scaling_factor(self.power, sf_current)

        if iscale.get_scaling_factor(self.electron_flow) is None:
            iscale.set_scaling_factor(self.electron_flow, sf_current * 1e5)

    @property
    def default_costing_method(self):
        return cost_electrolyzer



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    Var,
    assert_optimal_termination,
    units,
)
from pyomo.network import Port
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from watertap.unit_models.selective_oil_permeation import SelectiveOilPermeation
import watertap.property_models.selective_oil_permeation_prop_pack as props
from watertap.core.util.initialization import assert_no_degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent

from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    unscaled_variables_generator,
    badly_scaled_var_generator,
)
import idaes.logger as idaeslog


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.SopParameterBlock()
    m.fs.unit = SelectiveOilPermeation(property_package=m.fs.properties)

    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.is_isothermal
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.none
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.properties
    assert isinstance(m.fs.unit.feed_side.deltaP, Var)
    assert isinstance(m.fs.unit.deltaP, Var)


class TestSelectiveOilPermeation:
    @pytest.fixture(scope="class")
    def unit_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = props.SopParameterBlock()
        m.fs.unit = SelectiveOilPermeation(property_package=m.fs.properties)

        # fully specify system
        m.fs.unit.feed_side.properties_in[0].temperature.fix(298.15)  # temp in K
        m.fs.unit.feed_side.properties_in[0].pressure.fix(2.52 * units.bar)
        m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp["Liq", "H2O"].fix(
            6.27e-2
        )  # H2O flow in kg/s
        m.fs.unit.feed_side.properties_in[0].flow_mass_phase_comp["Liq", "oil"].fix(
            4.94e-4
        )  # oil flow in kg/s
        m.fs.unit.feed_side.properties_out[0].pressure.fix(2.24 * units.bar)
        m.fs.unit.area.fix(1.4)  # membrane area in m^2
        m.fs.unit.pore_diameter.fix(4.7e-8)
        m.fs.unit.porosity.fix(0.4)
        m.fs.unit.membrane_thickness.fix(4e-5)
        m.fs.unit.permeability_constant.fix(150)
        m.fs.unit.module_diameter.fix(0.064)
        m.fs.unit.properties_permeate[0].pressure.fix(1 * units.bar)

        return m

    @pytest.mark.unit
    def test_build(self, unit_frame):
        m = unit_frame

        # test ports
        port_lst = ["inlet", "retentate", "permeate"]
        for port_str in port_lst:
            port = getattr(m.fs.unit, port_str)
            assert (
                len(port.vars) == 3
            )  # number of state variables for SOP property package
            assert isinstance(port, Port)

        assert number_variables(m) == 35
        assert number_total_constraints(m) == 22
        assert number_unused_variables(m) == 0
        assert_units_consistent(m)

    @pytest.mark.unit
    def test_dof(self, unit_frame):
        m = unit_frame
        assert_no_degrees_of_freedom(m)

    @pytest.mark.unit
    def test_calculate_scaling(self, unit_frame):
        m = unit_frame

        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "H2O")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e5, index=("Liq", "oil")
        )
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0

    @pytest.mark.component
    def test_var_scaling(self, unit_frame):
        m = unit_frame

        # set initial values (to ensure these variables are not badly scaled)
        m.fs.unit.feed_side.properties_out[0].flow_mass_phase_comp[
            "Liq", "oil"
        ].value = 1e-5
        m.fs.unit.properties_permeate[0].flow_mass_phase_comp["Liq", "oil"].value = 1e-6
        m.fs.unit.feed_side.properties_in[0].vol_frac_phase_comp[
            "Liq", "oil"
        ].value = 1e-3

        badly_scaled_var_lst = list(
            badly_scaled_var_generator(m, large=1e2, small=1e-2, zero=1e-8)
        )
        assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_initialize(self, unit_frame):
        m = unit_frame
        m.fs.unit.initialize(outlvl=idaeslog.DEBUG)
        initialization_tester(unit_frame)

    @pytest.mark.component
    def test_solve(self, unit_frame):
        m = unit_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_conservation(self, unit_frame):
        m = unit_frame
        b = m.fs.unit
        comp_lst = m.fs.properties.component_list

        flow_mass_inlet = sum(
            b.feed_side.properties_in[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            b.feed_side.properties_out[0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            b.properties_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        assert (
            abs(value(flow_mass_inlet - flow_mass_retentate - flow_mass_permeate))
            <= 1e-6
        )

    @pytest.mark.component
    def test_solution(self, unit_frame):
        m = unit_frame
        assert pytest.approx(3.32237e-4, rel=1e-3) == value(
            m.fs.unit.properties_permeate[0].flow_mass_phase_comp["Liq", "oil"]
        )
        assert pytest.approx(1.38e5, rel=1e-3) == value(
            m.fs.unit.pressure_transmemb_avg[0]
        )
        assert pytest.approx(-2.8e4, rel=1e-3) == value(m.fs.unit.deltaP[0])
        assert pytest.approx(3.0425e-07, rel=1e-3) == value(m.fs.unit.flux_vol_oil[0])
        assert pytest.approx(3.3246e-06, rel=1e-3) == value(
            m.fs.unit.flux_vol_oil_pure[0]
        )
        assert pytest.approx(0.091512, rel=1e-3) == value(
            m.fs.unit.effective_area_ratio[0]
        )
        assert pytest.approx(0.10799, rel=1e-3) == value(
            m.fs.unit.effective_area_ratio_num[0]
        )
        assert pytest.approx(1.1801, rel=1e-3) == value(
            m.fs.unit.effective_area_ratio_den[0]
        )
        assert pytest.approx(0.67254, rel=1e-3) == value(m.fs.unit.recovery_frac_oil[0])
        assert pytest.approx(0.019687, rel=1e-3) == value(
            m.fs.unit.liquid_velocity_in[0]
        )
        assert pytest.approx(3.32237e-4, rel=1e-3) == value(
            m.fs.unit.mass_transfer_oil[0]
        )

    @pytest.mark.unit
    def test_report(self, unit_frame):
        unit_frame.fs.unit.report()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.common.config import Bool, ConfigBlock, ConfigValue, In
from pyomo.environ import (
    Var,
    check_optimal_termination,
    Suffix,
    NonNegativeReals,
    value,
    units as pyunits,
)

# Import IDAES cores
import idaes.logger as idaeslog
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.core.util.scaling as iscale

from watertap.core import ControlVolume0DBlock, InitializationMixin
from watertap.costing.unit_models.pressure_exchanger import cost_pressure_exchanger

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("PressureExchanger")
class PressureExchangerData(InitializationMixin, UnitModelBlockData):
    """
    Standard Pressure Exchanger Unit Model Class:
    - steady state only
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. Pressure exchangers do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. Pressure exchangers do not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "is_isothermal",
        ConfigValue(
            default=True,
            domain=Bool,
            description="""Assume isothermal conditions for control volume(s); energy_balance_type must be EnergyBalanceType.none,
    **default** - True.""",
        ),
    )

    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.none,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.none.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )

    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
        and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}""",
        ),
    )

    CONFIG.declare(
        "has_mass_transfer",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Defines if there is mass transport between high- and low-pressure sides",
            doc="""Indicates whether pressure exchanger solution mass transfer terms should be constructed or not.
    **default** - False.""",
        ),
    )

    def _validate_config(self):
        if (
            self.config.is_isothermal
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            raise ConfigurationError(
                "If the isothermal assumption is used then the energy balance type must be none"
            )

    def build(self):
        super().build()

        # Check configs for errors
        self._validate_config()

        # Pressure exchanger supports only liquid phase
        if self.config.property_package.phase_list != ["Liq"]:
            raise ConfigurationError(
                "Pressure exchanger model only supports one liquid phase ['Liq'],"
                "the property package has specified the following phases {}".format(
                    self.config.property_package.phase_list
                )
            )

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        units_meta = self.config.property_package.get_metadata().get_derived_units

        self.efficiency_pressure_exchanger = Var(
            self.flowsheet().config.time,
            initialize=0.95,
            bounds=(1e-6, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Pressure exchanger efficiency",
        )

        if self.config.has_mass_transfer:
            self.mass_transfer_fraction_comp = Var(
                self.flowsheet().config.time,
                self.config.property_package.component_list,
                initialize=0.05,
                bounds=(1e-6, 1),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="The fraction of solution transfering from high to low pressure side",
            )

        # Build control volume for high pressure side
        self.high_pressure_side = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.high_pressure_side.add_state_blocks(has_phase_equilibrium=False)

        self.high_pressure_side.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=self.config.has_mass_transfer,
        )

        self.high_pressure_side.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_enthalpy_transfer=False,
        )

        if self.config.is_isothermal:
            self.high_pressure_side.add_isothermal_assumption()

        self.high_pressure_side.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        @self.high_pressure_side.Expression(
            self.flowsheet().config.time,
            doc="Work transferred to high pressure side fluid (should be negative)",
        )
        def work(b, t):
            return b.properties_in[t].flow_vol * b.deltaP[t]

        # Build control volume for low pressure side
        self.low_pressure_side = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.low_pressure_side.add_state_blocks(has_phase_equilibrium=False)

        self.low_pressure_side.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=self.config.has_mass_transfer,
        )

        self.low_pressure_side.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_enthalpy_transfer=False,
        )

        if self.config.is_isothermal:
            self.low_pressure_side.add_isothermal_assumption()

        self.low_pressure_side.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        @self.low_pressure_side.Expression(
            self.flowsheet().config.time,
            doc="Work transferred to low pressure side fluid",
        )
        def work(b, t):  # pylint: disable=function-redefined
            return b.properties_in[t].flow_vol * b.deltaP[t]

        # Add Ports
        self.add_inlet_port(name="high_pressure_inlet", block=self.high_pressure_side)
        self.add_outlet_port(name="high_pressure_outlet", block=self.high_pressure_side)
        self.add_inlet_port(name="low_pressure_inlet", block=self.low_pressure_side)
        self.add_outlet_port(name="low_pressure_outlet", block=self.low_pressure_side)

        # Performance equations
        @self.Constraint(self.flowsheet().config.time, doc="Pressure transfer")
        def eq_pressure_transfer(b, t):
            return (
                b.low_pressure_side.deltaP[t]
                == b.efficiency_pressure_exchanger[t] * -b.high_pressure_side.deltaP[t]
            )

        @self.Constraint(self.flowsheet().config.time, doc="Equal volumetric flow rate")
        def eq_equal_flow_vol(b, t):

            return (
                b.high_pressure_side.properties_out[t].flow_vol
                == b.low_pressure_side.properties_in[t].flow_vol
            )

        @self.Constraint(
            self.flowsheet().config.time, doc="Equal low pressure on both sides"
        )
        def eq_equal_low_pressure(b, t):
            return (
                b.high_pressure_side.properties_out[t].pressure
                == b.low_pressure_side.properties_in[t].pressure
            )

        if self.config.has_mass_transfer:

            @self.Constraint(
                self.flowsheet().config.time,
                self.config.property_package.phase_list,
                self.config.property_package.component_list,
                doc="Mass transfer from high pressure side",
            )
            def eq_mass_transfer_from_high_pressure_side(b, t, p, j):
                comp = self.config.property_package.get_component(j)
                return b.high_pressure_side.mass_transfer_term[
                    t, p, j
                ] == -b.mass_transfer_fraction_comp[
                    t, j
                ] * b.high_pressure_side.properties_in[
                    t
                ].get_material_flow_terms(
                    p, j
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.config.property_package.phase_list,
                self.config.property_package.component_list,
                doc="Mass transfer term",
            )
            def eq_mass_transfer_term(b, t, p, j):
                return (
                    b.high_pressure_side.mass_transfer_term[t, p, j]
                    == -b.low_pressure_side.mass_transfer_term[t, p, j]
                )

    def initialize_build(
        self,
        state_args=None,
        routine=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure exchanger initialization routine

        Keyword Arguments:
            routine : str stating which initialization routine to execute
                        * None - currently no specialized routine for Pressure exchanger unit
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine (default=idaeslog.NOTSET)
            optarg : solver options dictionary object, if None provided an empty
                     dictionary will be used (default=None)
            solver : solver object or string indicating which solver to use during
                     initialization, if None provided the default solver will be used
                     (default = None)
        Returns: None
        """
        # Get loggers
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="properties")

        # Set solver and options
        opt = get_solver(solver, optarg)
        # initialize inlets
        flags_low_in = self.low_pressure_side.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=True,
        )
        flags_high_in = self.high_pressure_side.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=True,
        )

        init_log.info_high("Initialize inlets complete")

        # check that inlets are feasible
        if value(self.low_pressure_side.properties_in[0].pressure) > value(
            self.high_pressure_side.properties_in[0].pressure
        ):
            raise ConfigurationError(
                "Initializing pressure exchanger failed because "
                "the low pressure side inlet has a higher pressure "
                "than the high pressure side inlet"
            )
        # only needed when there is no mass trnasfer

        if (
            abs(
                value(self.low_pressure_side.properties_in[0].flow_vol)
                - value(self.high_pressure_side.properties_in[0].flow_vol)
            )
            / value(self.high_pressure_side.properties_in[0].flow_vol)
            > 1e-4
            and not self.config.has_mass_transfer
        ):  # flow_vol values are not within 0.1%
            raise ConfigurationError(
                "Initializing pressure exchanger failed because "
                "the volumetric flow rates are not equal for both inlets "
                + str(value(self.high_pressure_side.properties_out[0].flow_vol))
                + ","
                + str(value(self.low_pressure_side.properties_in[0].flow_vol))
            )
        else:  # volumetric flow is equal, deactivate flow constraint for the solve
            self.eq_equal_flow_vol.deactivate()

        # initialize outlets from inlets and update pressure
        def propogate_state(sb1, sb2):
            state_dict_1 = sb1.define_state_vars()
            state_dict_2 = sb2.define_state_vars()
            for k in state_dict_1.keys():
                if state_dict_1[k].is_indexed():
                    for m in state_dict_1[k].keys():
                        state_dict_2[k][m].value = state_dict_1[k][m].value
                else:
                    state_dict_2[k].value = state_dict_1[k].value

        # low pressure side
        propogate_state(
            self.low_pressure_side.properties_in[0],
            self.low_pressure_side.properties_out[0],
        )
        self.low_pressure_side.properties_out[
            0
        ].pressure = self.low_pressure_side.properties_in[
            0
        ].pressure.value + self.efficiency_pressure_exchanger[
            0
        ].value * (
            self.high_pressure_side.properties_in[0].pressure.value
            - self.low_pressure_side.properties_in[0].pressure.value
        )
        # high pressure side
        propogate_state(
            self.high_pressure_side.properties_in[0],
            self.high_pressure_side.properties_out[0],
        )
        self.high_pressure_side.properties_out[
            0
        ].pressure.value = self.low_pressure_side.properties_in[0].pressure.value
        init_log.info_high("Initialize outlets complete")

        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info("Initialization complete: {}".format(idaeslog.condition(res)))

        # release state of fixed variables
        self.low_pressure_side.properties_in.release_state(flags_low_in)
        self.high_pressure_side.properties_in.release_state(flags_high_in)

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

        # reactivate volumetric flow constraint
        self.eq_equal_flow_vol.activate()

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # scale variables
        if iscale.get_scaling_factor(self.efficiency_pressure_exchanger) is None:
            # efficiency should always be between 0.1-1
            iscale.set_scaling_factor(self.efficiency_pressure_exchanger, 1)
        if hasattr(self, "mass_transfer_fraction_comp"):
            if iscale.get_scaling_factor(self.mass_transfer_fraction_comp) is None:
                iscale.set_scaling_factor(self.mass_transfer_fraction_comp, 1)

        # scale expressions
        if iscale.get_scaling_factor(self.low_pressure_side.work) is None:
            sf = iscale.get_scaling_factor(
                self.low_pressure_side.properties_in[0].flow_vol
            )
            sf = sf * iscale.get_scaling_factor(self.low_pressure_side.deltaP[0])
            iscale.set_scaling_factor(self.low_pressure_side.work, sf)

        if iscale.get_scaling_factor(self.high_pressure_side.work) is None:
            sf = iscale.get_scaling_factor(
                self.high_pressure_side.properties_in[0].flow_vol
            )
            sf = sf * iscale.get_scaling_factor(self.high_pressure_side.deltaP[0])
            iscale.set_scaling_factor(self.high_pressure_side.work, sf)

        # transform constraints
        for t, c in self.eq_pressure_transfer.items():
            sf = iscale.get_scaling_factor(self.low_pressure_side.deltaP[t])
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_equal_flow_vol.items():
            sf = iscale.get_scaling_factor(
                self.low_pressure_side.properties_in[t].flow_vol
            )
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_equal_low_pressure.items():
            sf = iscale.get_scaling_factor(
                self.low_pressure_side.properties_in[t].pressure
            )
            iscale.constraint_scaling_transform(c, sf)

        if hasattr(self, "eq_mass_transfer_from_high_pressure_side"):
            for (t, p, j), c in self.eq_mass_transfer_from_high_pressure_side.items():
                sf = iscale.get_scaling_factor(
                    self.high_pressure_side.properties_in[t].get_material_flow_terms(
                        p, j
                    )
                )
                iscale.constraint_scaling_transform(c, sf)

        if hasattr(self, "eq_mass_transfer_term"):
            for (t, p, j), c in self.eq_mass_transfer_term.items():
                sf = iscale.get_scaling_factor(
                    self.high_pressure_side.mass_transfer_term[t, p, j]
                )
                iscale.constraint_scaling_transform(c, sf)
                sf = iscale.get_scaling_factor(
                    self.low_pressure_side.mass_transfer_term[t, p, j]
                )
                iscale.constraint_scaling_transform(c, sf)

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "HP Side In": self.high_pressure_inlet,
                "HP Side Out": self.high_pressure_outlet,
                "LP Side In": self.low_pressure_inlet,
                "LP Side Out": self.low_pressure_outlet,
            },
            time_point=time_point,
        )

    def _get_performance_contents(self, time_point=0):
        t = time_point
        return {
            "vars": {
                "Efficiency": self.efficiency_pressure_exchanger[t],
                "HP Side Pressure Change": self.high_pressure_side.deltaP[t],
                "LP Side Pressure Change": self.low_pressure_side.deltaP[t],
            },
            "exprs": {
                "HP Side Mechanical Work": self.high_pressure_side.work[t],
                "LP Side Mechanical Work": self.low_pressure_side.work[t],
            },
            "params": {},
        }

    @property
    def default_costing_method(self):
        return cost_pressure_exchanger



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import value
from watertap.examples.flowsheets.nf_dspmde.nf import main


@pytest.mark.component
def test_main():
    m = main()
    test_dict = {
        "lcow": [m.fs.costing.LCOW, 0.144580],
        "pressure": [m.fs.NF.pump.outlet.pressure[0] * 1e-5, 5.373047],
        "area": [m.fs.NF.nfUnit.area, 405.65321],
        "recovery": [
            m.fs.NF.nfUnit.recovery_vol_phase[0.0, "Liq"] * 100,
            89.999,
        ],
    }
    for model_result, testval in test_dict.values():
        assert pytest.approx(testval, rel=1e-3) == value(model_result)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import itertools

from pyomo.environ import (
    ConcreteModel,
    value,
    Param,
    Var,
    Constraint,
    Expression,
    Objective,
    TransformationFactory,
    Block,
    NonNegativeReals,
    RangeSet,
    Set,
    check_optimal_termination,
    units as pyunits,
)
from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock, UnitModelCostingBlock
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.initialization import propagate_state
from idaes.core.util.misc import StrEnum
from idaes.models.unit_models import Feed, Product, Mixer
from idaes.models.unit_models.mixer import MomentumMixingType
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslogger

from watertap.unit_models.reverse_osmosis_1D import (
    ReverseOsmosis1D,
    ConcentrationPolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
)
from watertap.unit_models.pressure_changer import Pump, EnergyRecoveryDevice
from watertap.core.util.initialization import (
    assert_no_degrees_of_freedom,
    assert_degrees_of_freedom,
)
from watertap.costing import (
    WaterTAPCosting,
    make_capital_cost_var,
    register_costing_parameter_block,
)
import watertap.property_models.NaCl_prop_pack as props


class ACase(StrEnum):
    fixed = "fixed"
    optimize = "optimize"
    single_optimum = "single_optimum"


class BCase(StrEnum):
    single_optimum = "single_optimum"
    optimize = "optimize"


class ABTradeoff(StrEnum):
    inequality_constraint = "inequality_constraint"
    equality_constraint = "equality_constraint"
    none = "none"


def run_lsrro_case(
    number_of_stages,
    water_recovery=None,
    Cin=None,
    Qin=None,
    Cbrine=None,
    A_case=ACase.fixed,
    B_case=BCase.optimize,
    AB_tradeoff=ABTradeoff.none,
    A_value=None,
    has_NaCl_solubility_limit=None,
    has_calculated_concentration_polarization=None,
    has_calculated_ro_pressure_drop=None,
    permeate_quality_limit=None,
    AB_gamma_factor=None,
    B_max=None,
    number_of_RO_finite_elements=10,
    set_default_bounds_on_module_dimensions=True,
):
    m = build(
        number_of_stages,
        has_NaCl_solubility_limit,
        has_calculated_concentration_polarization,
        has_calculated_ro_pressure_drop,
        number_of_RO_finite_elements,
        B_max,
    )
    set_operating_conditions(m, Cin, Qin)

    initialize(m)
    solve(m)
    print("\n***---Simulation results---***")
    display_system(m)
    display_design(m)
    display_state(m)

    optimize_set_up(
        m,
        set_default_bounds_on_module_dimensions,
        water_recovery,
        Cbrine,
        A_case,
        B_case,
        AB_tradeoff,
        A_value,
        permeate_quality_limit,
        AB_gamma_factor,
        B_max,
    )
    res = solve(m, raise_on_failure=False, tee=False)
    print("\n***---Optimization results---***")
    if check_optimal_termination(res):
        display_system(m)
        display_design(m)
        display_state(m)
        display_RO_reports(m)

    return m, res


def build(
    number_of_stages=2,
    has_NaCl_solubility_limit=True,
    has_calculated_concentration_polarization=True,
    has_calculated_ro_pressure_drop=True,
    number_of_RO_finite_elements=10,
    B_max=None,
):
    # ---building model---
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.NaClParameterBlock()
    m.fs.costing = WaterTAPCosting()

    m.fs.NumberOfStages = Param(initialize=number_of_stages)
    m.fs.Stages = RangeSet(m.fs.NumberOfStages)
    m.fs.LSRRO_Stages = RangeSet(2, m.fs.NumberOfStages)
    m.fs.NonFinalStages = RangeSet(m.fs.NumberOfStages - 1)
    if number_of_stages > 1:
        m.fs.IntermediateStages = RangeSet(2, m.fs.NumberOfStages - 1)
    else:
        m.fs.IntermediateStages = RangeSet(0)
    m.fs.FirstStage = m.fs.Stages.first()
    m.fs.LastStage = m.fs.Stages.last()

    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)

    # Add the mixers
    m.fs.Mixers = Mixer(
        m.fs.NonFinalStages,
        property_package=m.fs.properties,
        momentum_mixing_type=MomentumMixingType.equality,
        inlet_list=["upstream", "downstream"],
    )

    # Add the pumps
    m.fs.PrimaryPumps = Pump(m.fs.Stages, property_package=m.fs.properties)
    for pump in m.fs.PrimaryPumps.values():
        pump.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method=cost_high_pressure_pump_lsrro,
        )

    # Add the equalizer pumps
    m.fs.BoosterPumps = Pump(m.fs.LSRRO_Stages, property_package=m.fs.properties)
    for pump in m.fs.BoosterPumps.values():
        pump.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method=cost_high_pressure_pump_lsrro,
        )

    m.fs.total_pump_work = Expression(
        expr=sum(
            pyunits.convert(pump.work_mechanical[0], to_units=pyunits.kW)
            for pump in itertools.chain(
                m.fs.PrimaryPumps.values(), m.fs.BoosterPumps.values()
            )
        )
    )

    # Add the stages ROs
    if has_calculated_ro_pressure_drop:
        pressure_change_type = PressureChangeType.calculated
    else:
        pressure_change_type = PressureChangeType.fixed_per_stage

    if has_calculated_concentration_polarization:
        cp_type = ConcentrationPolarizationType.calculated
        kf_type = MassTransferCoefficient.calculated
    else:
        cp_type = ConcentrationPolarizationType.none
        kf_type = MassTransferCoefficient.none

    m.fs.ROUnits = ReverseOsmosis1D(
        m.fs.Stages,
        property_package=m.fs.properties,
        has_pressure_change=has_calculated_ro_pressure_drop,
        pressure_change_type=pressure_change_type,
        mass_transfer_coefficient=kf_type,
        concentration_polarization_type=cp_type,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=number_of_RO_finite_elements,
        has_full_reporting=True,
    )

    for idx, ro_stage in m.fs.ROUnits.items():
        if idx == m.fs.FirstStage:
            ro_stage.costing = UnitModelCostingBlock(
                flowsheet_costing_block=m.fs.costing,
                costing_method_arguments={"ro_type": "standard"},
            )
        else:
            ro_stage.costing = UnitModelCostingBlock(
                flowsheet_costing_block=m.fs.costing,
                costing_method_arguments={"ro_type": "high_pressure"},
            )

    # Add EnergyRecoveryDevices
    m.fs.EnergyRecoveryDeviceSet = Set(
        initialize=[m.fs.FirstStage, m.fs.LastStage]
        if m.fs.FirstStage < m.fs.LastStage
        else [m.fs.LastStage]
    )
    m.fs.EnergyRecoveryDevices = EnergyRecoveryDevice(
        m.fs.EnergyRecoveryDeviceSet, property_package=m.fs.properties
    )
    for erd in m.fs.EnergyRecoveryDevices.values():
        erd.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={
                "energy_recovery_device_type": "pressure_exchanger"
            },
        )

    m.fs.recovered_pump_work = Expression(
        expr=sum(
            pyunits.convert(erd.work_mechanical[0], to_units=pyunits.kW)
            for erd in m.fs.EnergyRecoveryDevices.values()
        )
    )
    m.fs.net_pump_work = Expression(
        expr=m.fs.total_pump_work + m.fs.recovered_pump_work
    )

    # additional parameters, variables or expressions ---------------------------------------------------------------------------
    m.fs.ro_min_pressure = Param(initialize=10e5, units=pyunits.Pa, mutable=True)
    m.fs.lsrro_min_pressure = Param(initialize=10e5, units=pyunits.Pa, mutable=True)
    m.fs.ro_max_pressure = Param(initialize=85e5, units=pyunits.Pa, mutable=True)
    m.fs.lsrro_max_pressure = Param(initialize=65e5, units=pyunits.Pa, mutable=True)

    if B_max is not None:
        m.fs.B_max = Param(initialize=B_max, mutable=True, units=pyunits.m / pyunits.s)

    # system water recovery
    m.fs.water_recovery = Var(
        initialize=0.5,
        bounds=(0, 1),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Water Recovery",
    )
    m.fs.eq_water_recovery = Constraint(
        expr=m.fs.feed.properties[0].flow_vol * m.fs.water_recovery
        == m.fs.product.properties[0].flow_vol
    )

    # costing and summary quantities

    # explicitly set the costing parameters used
    m.fs.costing.utilization_factor.fix(0.9)
    m.fs.costing.TIC.fix(2)
    m.fs.costing.factor_maintenance_labor_chemical.fix(0.03)
    m.fs.costing.factor_capital_annualization.fix(0.1)
    m.fs.costing.electricity_cost.set_value(0.07)
    m.fs.costing.reverse_osmosis.factor_membrane_replacement.fix(0.15)
    m.fs.costing.reverse_osmosis.membrane_cost.fix(30)
    m.fs.costing.reverse_osmosis.high_pressure_membrane_cost.fix(50)
    m.fs.costing.high_pressure_pump.cost.fix(53 / 1e5 * 3600)
    m.fs.costing.energy_recovery_device.pressure_exchanger_cost.fix(535)

    m.fs.costing.cost_process()

    product_flow_vol_total = m.fs.product.properties[0].flow_vol
    m.fs.costing.add_annual_water_production(product_flow_vol_total)
    m.fs.costing.add_specific_energy_consumption(product_flow_vol_total)
    m.fs.costing.add_LCOW(product_flow_vol_total)

    # objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # Expressions for parameter sweep -----------------------------------------
    # Final permeate concentration as mass fraction
    m.fs.product.properties[0].mass_frac_phase_comp

    # Touch feed concentration as mass concentration
    m.fs.feed.properties[0].conc_mass_phase_comp

    # Touch final brine concentration as mass concentration
    m.fs.disposal.properties[0].conc_mass_phase_comp

    # Touch final brine concentration as mass fraction
    m.fs.disposal.properties[0].mass_frac_phase_comp

    m.fs.mass_water_recovery = Expression(
        expr=m.fs.product.flow_mass_phase_comp[0, "Liq", "H2O"]
        / m.fs.feed.flow_mass_phase_comp[0, "Liq", "H2O"]
    )

    m.fs.system_salt_rejection = Expression(
        expr=1
        - m.fs.product.properties[0].conc_mass_phase_comp["Liq", "NaCl"]
        / m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "NaCl"]
    )
    m.fs.annual_feed = Expression(
        expr=pyunits.convert(
            m.fs.feed.properties[0].flow_vol, to_units=pyunits.m**3 / pyunits.year
        )
        * m.fs.costing.utilization_factor
    )
    m.fs.final_permeate_concentration = Expression(
        expr=m.fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"]
        / sum(m.fs.product.flow_mass_phase_comp[0, "Liq", j] for j in ["H2O", "NaCl"])
    )

    m.fs.costing.add_LCOW(m.fs.feed.properties[0].flow_vol, name="LCOW_feed")

    m.fs.costing.add_specific_energy_consumption(
        m.fs.feed.properties[0].flow_vol, name="specific_energy_consumption_feed"
    )

    @m.fs.Expression(m.fs.Stages)
    def stage_recovery_vol(fs, stage):
        return (
            fs.ROUnits[stage].mixed_permeate[0].flow_vol
            / fs.PrimaryPumps[stage].control_volume.properties_in[0].flow_vol
        )

    @m.fs.Expression(m.fs.Stages)
    def stage_recovery_mass_H2O(fs, stage):
        return (
            fs.ROUnits[stage].mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
            / m.fs.PrimaryPumps[stage]
            .control_volume.properties_in[0]
            .flow_mass_phase_comp["Liq", "H2O"]
        )

    m.fs.costing.primary_pump_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * sum(m.fs.PrimaryPumps[n].costing.direct_capital_cost for n in m.fs.Stages)
        / m.fs.costing.annual_water_production
    )

    m.fs.total_membrane_area = Expression(
        expr=sum(ro.area for ro in m.fs.ROUnits.values())
    )

    m.fs.costing.booster_pump_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * (
            sum(
                m.fs.BoosterPumps[n].costing.direct_capital_cost
                for n in m.fs.LSRRO_Stages
            )
            if number_of_stages > 1
            else 0 * m.fs.costing.base_currency
        )
        / m.fs.costing.annual_water_production,
    )

    m.fs.costing.erd_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * sum(
            erd.costing.direct_capital_cost
            for erd in m.fs.EnergyRecoveryDevices.values()
        )
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.electricity_lcow = Expression(
        expr=m.fs.costing.aggregate_flow_costs["electricity"]
        * m.fs.costing.utilization_factor
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.pumping_energy_aggregate_lcow = Expression(
        expr=m.fs.costing.factor_total_investment
        * m.fs.costing.TIC
        * (
            m.fs.costing.primary_pump_capex_lcow
            + (
                m.fs.costing.booster_pump_capex_lcow
                if number_of_stages > 1
                else 0 * m.fs.costing.base_currency / pyunits.m**3
            )
            + m.fs.costing.erd_capex_lcow
        )
        * (
            1
            + m.fs.costing.factor_maintenance_labor_chemical
            / m.fs.costing.factor_total_investment
            / m.fs.costing.factor_capital_annualization
        )
        + m.fs.costing.electricity_lcow
    )

    m.fs.costing.membrane_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * sum(m.fs.ROUnits[n].costing.direct_capital_cost for n in m.fs.Stages)
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.indirect_capex_lcow = Expression(
        expr=m.fs.costing.factor_capital_annualization
        * (m.fs.costing.total_capital_cost - m.fs.costing.aggregate_direct_capital_cost)
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.membrane_replacement_lcow = Expression(
        expr=sum(m.fs.ROUnits[n].costing.fixed_operating_cost for n in m.fs.Stages)
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.chemical_labor_maintenance_lcow = Expression(
        expr=m.fs.costing.maintenance_labor_chemical_operating_cost
        / m.fs.costing.annual_water_production
    )

    m.fs.costing.membrane_aggregate_lcow = Expression(
        expr=m.fs.costing.factor_total_investment
        * m.fs.costing.TIC
        * m.fs.costing.membrane_capex_lcow
        * (
            1
            + m.fs.costing.factor_maintenance_labor_chemical
            / m.fs.costing.factor_total_investment
            / m.fs.costing.factor_capital_annualization
        )
        + m.fs.costing.membrane_replacement_lcow
    )

    # Connections ------------------------------------------------------------
    # Connect the feed to the first pump
    m.fs.feed_to_pump = Arc(
        source=m.fs.feed.outlet, destination=m.fs.PrimaryPumps[1].inlet
    )

    # Connect the primary RO permeate to the product
    m.fs.primary_RO_to_product = Arc(
        source=m.fs.ROUnits[1].permeate, destination=m.fs.product.inlet
    )

    # Connect the Pump n to the Mixer n
    m.fs.pump_to_mixer = Arc(
        m.fs.NonFinalStages,
        rule=lambda fs, n: {
            "source": fs.PrimaryPumps[n].outlet,
            "destination": fs.Mixers[n].upstream,
        },
    )

    # Connect the Mixer n to the Stage n
    m.fs.mixer_to_stage = Arc(
        m.fs.NonFinalStages,
        rule=lambda fs, n: {
            "source": fs.Mixers[n].outlet,
            "destination": fs.ROUnits[n].inlet,
        },
    )

    # Connect the Stage n to the Eq Pump n
    m.fs.stage_permeate_to_booster_pump = Arc(
        m.fs.LSRRO_Stages,
        rule=lambda fs, n: {
            "source": fs.ROUnits[n].permeate,
            "destination": fs.BoosterPumps[n].inlet,
        },
    )

    # Connect the Eq Pump n to the Mixer n-1
    m.fs.booster_pump_to_mixer = Arc(
        m.fs.LSRRO_Stages,
        rule=lambda fs, n: {
            "source": fs.BoosterPumps[n].outlet,
            "destination": fs.Mixers[n - 1].downstream,
        },
    )

    last_stage = m.fs.LastStage
    if number_of_stages > 1:
        # Connect the primary RO permeate to the product
        m.fs.primary_RO_to_erd = Arc(
            source=m.fs.ROUnits[1].retentate,
            destination=m.fs.EnergyRecoveryDevices[1].inlet,
        )
        # Connect 1st stage ERD to primary pump
        m.fs.primary_ERD_to_pump = Arc(
            source=m.fs.EnergyRecoveryDevices[1].outlet,
            destination=m.fs.PrimaryPumps[2].inlet,
        )

    # Connect the Stage n to the Pump n+1
    m.fs.stage_retentate_to_pump = Arc(
        m.fs.IntermediateStages,
        rule=lambda fs, n: {
            "source": fs.ROUnits[n].retentate,
            "destination": fs.PrimaryPumps[n + 1].inlet,
        },
    )
    # Connect the Pump N to the Stage N
    m.fs.pumpN_to_stageN = Arc(
        source=m.fs.PrimaryPumps[last_stage].outlet,
        destination=m.fs.ROUnits[last_stage].inlet,
    )
    # Connect Final Stage to EnergyRecoveryDevice Pump
    m.fs.stage_to_erd = Arc(
        source=m.fs.ROUnits[last_stage].retentate,
        destination=m.fs.EnergyRecoveryDevices[last_stage].inlet,
    )
    # Connect the EnergyRecoveryDevice to the disposal
    m.fs.erd_to_disposal = Arc(
        source=m.fs.EnergyRecoveryDevices[last_stage].outlet,
        destination=m.fs.disposal.inlet,
    )

    # additional bounding
    if has_NaCl_solubility_limit:
        for b in m.component_data_objects(Block, descend_into=True):
            # NaCl solubility limit
            if hasattr(b, "is_property_constructed") and b.is_property_constructed(
                "mass_frac_phase_comp"
            ):
                b.mass_frac_phase_comp["Liq", "NaCl"].setub(0.2614)

    TransformationFactory("network.expand_arcs").apply_to(m)

    return m


def build_high_pressure_pump_cost_param_block(blk):
    blk.cost = Var(
        initialize=53 / 1e5 * 3600,
        doc="High pressure pump cost",
        units=pyunits.USD_2018 / pyunits.watt,
    )


@register_costing_parameter_block(
    build_rule=build_high_pressure_pump_cost_param_block,
    parameter_block_name="high_pressure_pump",
)
def cost_high_pressure_pump_lsrro(blk, cost_electricity_flow=True):
    t0 = blk.flowsheet().time.first()
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * (
            blk.costing_package.high_pressure_pump.cost
            * pyunits.watt
            / (pyunits.m**3 * pyunits.pascal / pyunits.s)
            * blk.unit_model.outlet.pressure[t0]
            * blk.unit_model.control_volume.properties_out[t0].flow_vol
        )
    )

    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyunits.convert(blk.unit_model.work_mechanical[t0], to_units=pyunits.kW),
            "electricity",
        )


def set_operating_conditions(m, Cin=None, Qin=None):
    # ---specifications---
    if Qin is None:
        Qin = 1e-3
    # parameters
    pump_efi = 0.75  # pump efficiency [-]
    erd_efi = 0.8  # energy recovery device efficiency [-]
    mem_A = 4.2e-12  # membrane water permeability coefficient [m/s-Pa]
    mem_B = 3.5e-8  # membrane salt permeability coefficient [m/s]
    height = 1e-3  # channel height in membrane stage [m]
    spacer_porosity = 0.85  # spacer porosity in membrane stage [-]
    width = 5 * Qin / 1e-3  # effective membrane width [m]
    area = 100 * Qin / 1e-3  # membrane area [m^2]
    pressure_atm = 101325  # atmospheric pressure [Pa]

    # feed
    # feed_flow_mass = 1*pyunits.kg/pyunits.s
    if Cin is None:
        Cin = 70

    feed_temperature = 273.15 + 20

    # initialize feed
    m.fs.feed.pressure[0].fix(pressure_atm)
    m.fs.feed.temperature[0].fix(feed_temperature)
    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "NaCl"] = Cin
    m.fs.feed.properties.calculate_state(
        var_args={
            ("conc_mass_phase_comp", ("Liq", "NaCl")): value(
                m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "NaCl"]
            ),  # feed mass concentration
            ("flow_vol_phase", "Liq"): Qin,
        },  # volumetric feed flowrate [-]
        hold_state=True,  # fixes the calculated component mass flow rates
    )

    # initialize pumps
    for pump in m.fs.PrimaryPumps.values():
        pump.control_volume.properties_out[0].pressure = 75e5
        pump.efficiency_pump.fix(pump_efi)
        pump.control_volume.properties_out[0].pressure.fix()
        iscale.set_scaling_factor(pump.control_volume.work, 1e-3)

    # initialize eq pumps
    for pump in m.fs.BoosterPumps.values():
        pump.efficiency_pump.fix(pump_efi)
        iscale.set_scaling_factor(pump.control_volume.work, 1e-3)

    # initialize stages
    for idx, stage in m.fs.ROUnits.items():
        if idx == m.fs.FirstStage:
            B_scale = 1.0
        else:
            B_scale = 100.0

        stage.A_comp.fix(mem_A)
        stage.B_comp.fix(mem_B * B_scale)
        stage.area.fix(area / float(idx))
        stage.width.fix(width)
        stage.mixed_permeate[0].pressure.fix(pressure_atm)
        if (
            stage.config.mass_transfer_coefficient == MassTransferCoefficient.calculated
        ) or stage.config.pressure_change_type == PressureChangeType.calculated:
            stage.feed_side.channel_height.fix(height)
            stage.feed_side.spacer_porosity.fix(spacer_porosity)

    # energy recovery devices
    for erd in m.fs.EnergyRecoveryDevices.values():
        erd.efficiency_pump.fix(erd_efi)
        iscale.set_scaling_factor(erd.control_volume.work, 1e-3)

    # if FirstStage *is* LastStage, we'll just overwrite the pressure
    m.fs.EnergyRecoveryDevices[m.fs.FirstStage].control_volume.properties_out[
        0
    ].pressure.fix(70e5)
    m.fs.EnergyRecoveryDevices[m.fs.LastStage].control_volume.properties_out[
        0
    ].pressure.fix(pressure_atm)

    # ---scaling---
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "NaCl")
    )
    iscale.calculate_scaling_factors(m)

    # ---checking model---
    assert_units_consistent(m)
    assert_no_degrees_of_freedom(m)

    print(
        "Feed Concentration = %.1f ppt"
        % (value(m.fs.feed.flow_mass_phase_comp[0, "Liq", "NaCl"]) * 1000)
    )


def _lsrro_mixer_guess_initializer(
    mixer, solvent_multiplier, solute_multiplier, optarg
):

    for vname in mixer.upstream.vars:
        if vname == "flow_mass_phase_comp":
            for time, phase, comp in mixer.upstream.vars[vname]:
                if comp in mixer.config.property_package.solute_set:
                    mixer.downstream.vars[vname][time, phase, comp].value = (
                        solute_multiplier
                        * mixer.upstream.vars[vname][time, phase, comp].value
                    )
                elif comp in mixer.config.property_package.solvent_set:
                    mixer.downstream.vars[vname][time, phase, comp].value = (
                        solvent_multiplier
                        * mixer.upstream.vars[vname][time, phase, comp].value
                    )
                else:
                    raise RuntimeError(f"Unknown component {comp}")
        else:  # copy the state
            for idx in mixer.upstream.vars[vname]:
                mixer.downstream.vars[vname][idx].value = mixer.upstream.vars[vname][
                    idx
                ].value

    mixer.initialize(optarg=optarg)


def do_forward_initialization_pass(m, optarg, guess_mixers):
    print("--------------------START FORWARD INITIALIZATION PASS--------------------")
    # start with the feed
    m.fs.feed.initialize(optarg=optarg)

    propagate_state(m.fs.feed_to_pump)

    last_stage = m.fs.LastStage
    first_stage = m.fs.FirstStage

    for stage in m.fs.Stages:
        m.fs.PrimaryPumps[stage].initialize(optarg=optarg)

        if stage == last_stage:
            propagate_state(m.fs.pumpN_to_stageN)
        else:
            propagate_state(m.fs.pump_to_mixer[stage])
            if guess_mixers:
                _lsrro_mixer_guess_initializer(
                    m.fs.Mixers[stage],
                    solvent_multiplier=0.5,
                    solute_multiplier=0.2,
                    optarg=optarg,
                )
            else:
                m.fs.Mixers[stage].initialize(optarg=optarg)
            propagate_state(m.fs.mixer_to_stage[stage])

        try:
            m.fs.ROUnits[stage].initialize(optarg=optarg)
        except InitializationError:
            pass

        if stage == first_stage:
            propagate_state(m.fs.primary_RO_to_product)
            m.fs.product.initialize(optarg=optarg)
            if value(m.fs.NumberOfStages) > 1:
                propagate_state(m.fs.primary_RO_to_erd)
                m.fs.EnergyRecoveryDevices[first_stage].initialize(optarg=optarg)
                propagate_state(m.fs.primary_ERD_to_pump)
        else:
            propagate_state(m.fs.stage_permeate_to_booster_pump[stage])
            m.fs.BoosterPumps[stage].initialize(optarg=optarg)
            propagate_state(m.fs.booster_pump_to_mixer[stage])

        if stage in m.fs.IntermediateStages:
            propagate_state(m.fs.stage_retentate_to_pump[stage])

    # for the end stage
    propagate_state(m.fs.stage_to_erd)
    m.fs.EnergyRecoveryDevices[last_stage].initialize(optarg=optarg)
    propagate_state(m.fs.erd_to_disposal)
    m.fs.disposal.initialize(optarg=optarg)


def do_backward_initialization_pass(m, optarg):
    print("--------------------START BACKWARD INITIALIZATION PASS--------------------")

    first_stage = m.fs.FirstStage
    for stage in reversed(m.fs.NonFinalStages):
        m.fs.Mixers[stage].initialize(optarg=optarg)
        propagate_state(m.fs.mixer_to_stage[stage])
        try:
            m.fs.ROUnits[stage].initialize(optarg=optarg)
        except InitializationError:
            pass
        if stage == first_stage:
            if value(m.fs.NumberOfStages) > 1:
                propagate_state(m.fs.primary_ERD_to_pump)
                m.fs.EnergyRecoveryDevices[first_stage].initialize(optarg=optarg)
                propagate_state(m.fs.primary_RO_to_erd)
            propagate_state(m.fs.primary_RO_to_product)
            m.fs.product.initialize(optarg=optarg)
        else:
            propagate_state(m.fs.stage_retentate_to_pump[stage])
            propagate_state(m.fs.stage_permeate_to_booster_pump[stage])
            m.fs.BoosterPumps[stage].initialize(optarg=optarg)
            propagate_state(m.fs.booster_pump_to_mixer[stage])


def initialize(m, verbose=True, solver=None):

    # ---initializing---
    # set up solvers
    if solver is None:
        solver = get_solver()

    optarg = solver.options
    do_forward_initialization_pass(m, optarg=optarg, guess_mixers=True)
    for _ in range(m.fs.NumberOfStages.value // 2):
        do_backward_initialization_pass(m, optarg=optarg)
        do_forward_initialization_pass(m, optarg=optarg, guess_mixers=False)

    # # set up SD tool
    seq = SequentialDecomposition()
    seq.options.tear_method = "Direct"
    seq.options.iterLim = m.fs.NumberOfStages
    seq.options.tear_set = list(m.fs.booster_pump_to_mixer.values())
    seq.options.log_info = True

    # run SD tool
    def func_initialize(unit):
        outlvl = idaeslogger.INFO if verbose else idaeslogger.CRITICAL
        try:
            unit.initialize(optarg=solver.options, outlvl=outlvl)
        except InitializationError:
            pass

    seq.run(m, func_initialize)

    m.fs.costing.initialize()


def solve(model, solver=None, tee=False, raise_on_failure=False):
    # ---solving---
    if solver is None:
        solver = get_solver()

    results = solver.solve(model, tee=tee)
    if check_optimal_termination(results):
        return results
    msg = (
        "The current configuration is infeasible. Please adjust the decision variables."
    )
    if raise_on_failure:
        raise RuntimeError(msg)
    else:
        print(msg)
        return results


def optimize_set_up(
    m,
    set_default_bounds_on_module_dimensions,
    water_recovery=None,
    Cbrine=None,
    A_case=ACase.fixed,
    B_case=BCase.optimize,
    AB_tradeoff=ABTradeoff.none,
    A_value=None,
    permeate_quality_limit=None,
    AB_gamma_factor=None,
    B_max=None,
):
    """
    Get the LSRRO flowsheet ready to optimize

    Attributes
    ----------
    B_case: 'single_optimum' or anything else to optimize B value at every LSR stage

    A_case: 'fixed' or 'optimize' or 'single_optimum' A at every LSR stage

    AB_tradeoff: 'inequality_constraint' B >= function of A
                 'equality_constraint' B = function of A
                 'none' no constraint relating B value to A value

    A_value: if A_case='fixed', then provide a value to fix A with

    Returns
    -------
    model (Pyomo ConcreteModel) : The LSRRO flowsheet.

    """

    for idx, pump in m.fs.PrimaryPumps.items():
        pump.control_volume.properties_out[0].pressure.unfix()
        pump.deltaP.setlb(0)
        if idx > m.fs.Stages.first():
            pump.max_lsrro_pressure_con = Constraint(
                expr=(
                    m.fs.lsrro_min_pressure,
                    pump.control_volume.properties_out[0].pressure,
                    m.fs.lsrro_max_pressure,
                )
            )
            iscale.constraint_scaling_transform(
                pump.max_lsrro_pressure_con,
                iscale.get_scaling_factor(
                    pump.control_volume.properties_out[0].pressure
                ),
            )
        else:
            pump.max_ro_pressure_con = Constraint(
                expr=(
                    m.fs.ro_min_pressure,
                    pump.control_volume.properties_out[0].pressure,
                    m.fs.ro_max_pressure,
                )
            )
            iscale.constraint_scaling_transform(
                pump.max_ro_pressure_con,
                iscale.get_scaling_factor(
                    pump.control_volume.properties_out[0].pressure
                ),
            )

    if value(m.fs.NumberOfStages) > 1:
        m.fs.EnergyRecoveryDevices[1].control_volume.properties_out[0].pressure.unfix()
        m.fs.EnergyRecoveryDevices[1].deltaP.setub(0)

    # unfix eq pumps
    for idx, pump in m.fs.BoosterPumps.items():
        pump.control_volume.properties_out[0].pressure.unfix()
        pump.max_ro_pressure_con = Constraint(
            expr=(
                m.fs.ro_min_pressure,
                pump.control_volume.properties_out[0].pressure,
                m.fs.ro_max_pressure,
            )
        )
        iscale.constraint_scaling_transform(
            pump.max_ro_pressure_con,
            iscale.get_scaling_factor(pump.control_volume.properties_out[0].pressure),
        )
        pump.deltaP.setlb(0)

    if B_case == BCase.single_optimum:
        m.fs.B_comp_system = Var(
            domain=NonNegativeReals,
            units=pyunits.m * pyunits.s**-1,
            doc="Solute permeability coeff. constant in all LSR stages",
        )
        m.fs.B_comp_system.set_value(
            m.fs.ROUnits[m.fs.LSRRO_Stages.first()].B_comp[0, "NaCl"]
        )
        m.fs.B_comp_system.setlb(3.5e-8)
        if B_max is None:
            m.fs.B_comp_system.setub(3.5e-8 * 1e2)
        else:
            m.fs.B_comp_system.setub(m.fs.B_max)
    if A_case == ACase.single_optimum:
        m.fs.A_comp_system = Var(
            domain=NonNegativeReals,
            units=pyunits.m * pyunits.s**-1 * pyunits.Pa**-1,
            doc="Water permeability coeff. constant in all LSR stages",
        )
        m.fs.A_comp_system.set_value(m.fs.ROUnits[2].A_comp[0, "H2O"])
        m.fs.A_comp_system.setlb(2.78e-12)
        m.fs.A_comp_system.setub(4.2e-11)
    if (
        AB_tradeoff == ABTradeoff.equality_constraint
        or AB_tradeoff == ABTradeoff.inequality_constraint
    ):
        m.fs.AB_tradeoff_coeff = Param(initialize=0.01333, mutable=True)
        if AB_gamma_factor is None:
            AB_gamma_factor = 1
        m.fs.AB_tradeoff_coeff.set_value(
            AB_gamma_factor * value(m.fs.AB_tradeoff_coeff)
        )

    # unfix stages
    for idx, stage in m.fs.ROUnits.items():
        stage.area.unfix()
        stage.width.unfix()
        if set_default_bounds_on_module_dimensions:
            # bounds originally set for Cost Optimization of LSRRO paper
            stage.area.setlb(1)
            stage.area.setub(20000)
            stage.width.setlb(0.1)
            stage.width.setub(1000)
        else:
            stage.area.setlb(1)
            stage.area.setub(None)
            stage.width.setlb(1)
            stage.width.setub(None)
            stage.length.setlb(1)

        if (
            stage.config.mass_transfer_coefficient == MassTransferCoefficient.calculated
        ) or (stage.config.pressure_change_type == PressureChangeType.calculated):
            stage.feed_side.N_Re[0, 0].unfix()

        if idx > m.fs.Stages.first():
            stage.B_comp.unfix()
            stage.B_comp.setlb(3.5e-8)
            if B_max is not None:
                stage.B_comp.setub(m.fs.B_max)
            else:
                stage.B_comp.setub(None)
            if B_case == BCase.single_optimum:
                stage.B_comp_equal = Constraint(
                    expr=stage.B_comp[0, "NaCl"] == m.fs.B_comp_system
                )
            if A_case == ACase.single_optimum:
                stage.A_comp_equal = Constraint(
                    expr=stage.A_comp[0, "H2O"] == m.fs.A_comp_system
                )

            stage.A_min = Param(
                initialize=2.78e-12, units=pyunits.m**2 / pyunits.kg * pyunits.s
            )
            stage.A_max = Param(
                initialize=4.2e-11, units=pyunits.m**2 / pyunits.kg * pyunits.s
            )
            stage._A_comp_con = Constraint(
                expr=(stage.A_min, stage.A_comp[0, "H2O"], stage.A_max)
            )
            iscale.constraint_scaling_transform(
                stage._A_comp_con, iscale.get_scaling_factor(stage.A_comp[0, "H2O"])
            )
            if A_case == ACase.optimize or A_case == ACase.single_optimum:
                stage.A_comp.unfix()
                stage.A_comp.setlb(2.78e-12)
                stage.A_comp.setub(4.2e-11)
            elif A_case == ACase.fixed:
                if not isinstance(A_value, (int, float)):
                    raise TypeError("A_value must be a numeric value")
                stage.A_comp.unfix()
                stage.A_comp.fix(A_value)
            else:
                raise TypeError(
                    f'A_case must be set to "fix", "single_optimum", "optimize" or None.'
                    f" A_case was set to {A_case}"
                )

            if AB_tradeoff == ABTradeoff.equality_constraint:
                stage.ABtradeoff = Constraint(
                    expr=pyunits.convert(
                        stage.B_comp[0, "NaCl"],
                        to_units=pyunits.L * pyunits.m**-2 * pyunits.hour**-1,
                    )
                    == m.fs.AB_tradeoff_coeff
                    * pyunits.convert(
                        stage.A_comp[0, "H2O"],
                        to_units=pyunits.L
                        * pyunits.m**-2
                        * pyunits.bar**-1
                        * pyunits.hour**-1,
                    )
                    ** 3
                    * pyunits.L**-2
                    * pyunits.m**4
                    * pyunits.hour**2
                    * pyunits.bar**3
                )
            elif AB_tradeoff == ABTradeoff.inequality_constraint:
                stage.ABtradeoff = Constraint(
                    expr=pyunits.convert(
                        stage.B_comp[0, "NaCl"],
                        to_units=pyunits.L * pyunits.m**-2 * pyunits.hour**-1,
                    )
                    >= m.fs.AB_tradeoff_coeff
                    * pyunits.convert(
                        stage.A_comp[0, "H2O"],
                        to_units=pyunits.L
                        * pyunits.m**-2
                        * pyunits.bar**-1
                        * pyunits.hour**-1,
                    )
                    ** 3
                    * pyunits.L**-2
                    * pyunits.m**4
                    * pyunits.hour**2
                    * pyunits.bar**3
                )
            else:
                pass

    min_avg_flux = 1  # minimum average water flux [kg/m2-h]
    # [kg/m2-s]
    min_avg_flux = min_avg_flux / 3600 * pyunits.kg / pyunits.m**2 / pyunits.s

    # additional constraints
    if water_recovery is not None:
        # product mass flow rate fraction of feed [-]
        m.fs.water_recovery.fix(water_recovery)
    if Cbrine is not None:
        # Final brine concentration
        m.fs.ROUnits[m.fs.Stages.last()].feed_side.properties[
            0, 1
        ].conc_mass_phase_comp["Liq", "NaCl"].fix(Cbrine)

    # add upper bound for permeate concentration
    if permeate_quality_limit is not None:
        if isinstance(permeate_quality_limit, (int, float)):
            m.fs.ROUnits[1].mixed_permeate[0].mass_frac_phase_comp["Liq", "NaCl"].setub(
                permeate_quality_limit
            )
        else:
            raise TypeError("permeate_quality_limit must be None, integer, or float")
    # ---checking model---
    assert_units_consistent(m)

    assert_degrees_of_freedom(
        m,
        4 * m.fs.NumberOfStages
        - (1 if (water_recovery is not None) else 0)
        - (1 if value(m.fs.NumberOfStages) == 1 else 0),
    )

    return m


def display_design(m):
    print("--decision variables--")
    for stage in m.fs.Stages:
        print(
            "Stage %d operating pressure %.1f bar"
            % (stage, m.fs.ROUnits[stage].inlet.pressure[0].value / 1e5)
        )
        print(
            "Stage %d membrane area      %.1f m2"
            % (stage, m.fs.ROUnits[stage].area.value)
        )
        print(
            "Stage %d water perm. coeff.  %.1f LMH/bar"
            % (stage, m.fs.ROUnits[stage].A_comp[0, "H2O"].value * (3.6e11))
        )
        print(
            "Stage %d salt perm. coeff.  %.1f LMH"
            % (stage, m.fs.ROUnits[stage].B_comp[0, "NaCl"].value * (1000.0 * 3600.0))
        )


def display_state(m):
    print("--------state---------")

    def print_state(s, b):
        flow_mass = sum(
            b.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
        )
        mass_frac_ppm = b.flow_mass_phase_comp[0, "Liq", "NaCl"].value / flow_mass * 1e6
        pressure_bar = b.pressure[0].value / 1e5
        print(
            s.ljust(20)
            + ": %.3f kg/s, %.0f ppm, %.1f bar"
            % (flow_mass, mass_frac_ppm, pressure_bar)
        )

    print_state("Feed", m.fs.feed.outlet)
    for stage in m.fs.Stages:

        print_state(f"Primary Pump {stage} out", m.fs.PrimaryPumps[stage].outlet)
        if stage == m.fs.LastStage:
            pass
        else:
            print_state(f"Mixer {stage} recycle", m.fs.Mixers[stage].downstream)
            print_state(f"Mixer {stage} out", m.fs.Mixers[stage].outlet)

        print_state(f"RO {stage} permeate", m.fs.ROUnits[stage].permeate)
        print_state(f"RO {stage} retentate", m.fs.ROUnits[stage].retentate)
        wr = m.fs.ROUnits[stage].recovery_vol_phase[0, "Liq"].value

        if stage == m.fs.FirstStage:
            pass
        else:
            print_state(f"Booster Pump {stage} out", m.fs.BoosterPumps[stage].outlet)

    print_state(f"Disposal", m.fs.disposal.inlet)
    print_state(f"Product", m.fs.product.inlet)


def display_system(m):
    print("----system metrics----")
    feed_flow_mass = sum(
        m.fs.feed.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    feed_mass_frac_NaCl = (
        m.fs.feed.flow_mass_phase_comp[0, "Liq", "NaCl"].value / feed_flow_mass
    )
    print("Feed: %.2f kg/s, %.0f ppm" % (feed_flow_mass, feed_mass_frac_NaCl * 1e6))

    prod_flow_mass = sum(
        m.fs.product.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    prod_mass_frac_NaCl = (
        m.fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"].value / prod_flow_mass
    )
    print("Product: %.3f kg/s, %.0f ppm" % (prod_flow_mass, prod_mass_frac_NaCl * 1e6))

    brine_flow_mass = sum(
        m.fs.disposal.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    brine_mass_frac_NaCl = (
        m.fs.disposal.flow_mass_phase_comp[0, "Liq", "NaCl"].value / brine_flow_mass
    )
    print("Brine: %.3f kg/s, %.0f ppm" % (brine_flow_mass, brine_mass_frac_NaCl * 1e6))

    print("Volumetric water recovery: %.1f%%" % (value(m.fs.water_recovery) * 100))
    print(f"Number of Stages: {value(m.fs.NumberOfStages)}")
    total_area = value(sum(m.fs.ROUnits[a].area for a in m.fs.Stages))
    print(f"Total Membrane Area: {total_area:.2f}")
    print(
        "Energy Consumption: %.1f kWh/m3"
        % value(m.fs.costing.specific_energy_consumption)
    )

    print("Levelized cost of water: %.2f $/m3" % value(m.fs.costing.LCOW))
    print(
        f"Primary Pump Capital Cost ($/m3):"
        f"{value(m.fs.costing.factor_capital_annualization*sum(m.fs.PrimaryPumps[stage].costing.capital_cost for stage in m.fs.Stages)/ m.fs.costing.annual_water_production)}"
    )
    print(
        f"Booster Pump Capital Cost ($/m3): "
        f"{value(m.fs.costing.factor_capital_annualization*sum(m.fs.BoosterPumps[stage].costing.capital_cost for stage in m.fs.LSRRO_Stages) / m.fs.costing.annual_water_production)}"
    )
    print(
        f"ERD Capital Cost ($/m3):"
        f"{value(m.fs.costing.factor_capital_annualization*sum(erd.costing.capital_cost for erd in m.fs.EnergyRecoveryDevices.values()) / m.fs.costing.annual_water_production)}"
    )
    print(
        f"Membrane Capital Cost ($/m3): "
        f"{value(m.fs.costing.factor_capital_annualization*sum(m.fs.ROUnits[stage].costing.capital_cost for stage in m.fs.Stages) / m.fs.costing.annual_water_production)}"
    )
    print(
        f"Indirect Capital Cost ($/m3): "
        f"{value(m.fs.costing.factor_capital_annualization*(m.fs.costing.total_capital_cost - m.fs.costing.aggregate_capital_cost) / m.fs.costing.annual_water_production)}"
    )
    electricity_cost = value(
        m.fs.costing.aggregate_flow_costs["electricity"]
        * m.fs.costing.utilization_factor
        / m.fs.costing.annual_water_production
    )
    print(f"Electricity cost ($/m3): {electricity_cost}")

    print("\n")


def display_RO_reports(m):
    for stage in m.fs.ROUnits.values():
        stage.report()


if __name__ == "__main__":
    m, results = run_lsrro_case(
        number_of_stages=3,
        water_recovery=0.50,
        Cin=70,  # inlet NaCl conc kg/m3,
        Qin=1e-3,  # inlet feed flowrate m3/s
        Cbrine=None,  # brine conc kg/m3
        A_case=ACase.optimize,
        B_case=BCase.optimize,
        AB_tradeoff=ABTradeoff.equality_constraint,
        # A_value=4.2e-12, #membrane water permeability coeff m/s-Pa
        has_NaCl_solubility_limit=True,
        has_calculated_concentration_polarization=True,
        has_calculated_ro_pressure_drop=True,
        permeate_quality_limit=500e-6,
        AB_gamma_factor=1,
        B_max=3.5e-6,
        number_of_RO_finite_elements=10,
        set_default_bounds_on_module_dimensions=True,
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a hydrothermal gasification unit.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("HTGZO")
class HTGZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a hydrothermal gasification (HTG) unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "hydrothermal_gasification"

        build_sido_reactive(self)

        self.flow_mass_in = Var(
            self.flowsheet().time,
            units=pyunits.t / pyunits.hour,
            bounds=(0, None),
            doc="Inlet mass flowrate",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for inlet mass flowrate.",
        )
        def cons_flow_mass(b, t):
            return b.flow_mass_in[t] == pyunits.convert(
                sum(
                    b.properties_in[t].flow_mass_comp[j]
                    for j in b.properties_in[t].component_list
                ),
                to_units=pyunits.t / pyunits.hour,
            )

        self._perf_var_dict["Inlet Mass Flowrate"] = self.flow_mass_in

        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity consumption of unit",
        )

        self._perf_var_dict["Electricity Demand"] = self.electricity

        self.energy_electric_flow_mass = Var(
            units=pyunits.kWh / pyunits.t,
            doc="Electricity intensity with respect to inlet flowrate",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on inlet flowrate.",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == pyunits.convert(
                b.energy_electric_flow_mass * b.flow_mass_in[t], to_units=pyunits.kW
            )

        self._fixed_perf_vars.append(self.energy_electric_flow_mass)
        self._perf_var_dict["Electricity Intensity"] = self.energy_electric_flow_mass

        self.catalyst_dosage = Var(
            units=pyunits.pound / pyunits.t,
            bounds=(0, None),
            doc="Dosage of catalyst per inlet flow",
        )

        self._fixed_perf_vars.append(self.catalyst_dosage)

        self._perf_var_dict["Dosage of catalyst per inlet flow"] = self.catalyst_dosage

        self.catalyst_flow = Var(
            self.flowsheet().time,
            units=pyunits.pound / pyunits.hr,
            bounds=(0, None),
            doc="Catalyst flow",
        )

        self._perf_var_dict["Catalyst flow"] = self.catalyst_flow

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for catalyst flow based on inlet flow rate.",
        )
        def eq_catalyst_flow(b, t):
            return b.catalyst_flow[t] == pyunits.convert(
                b.catalyst_dosage * b.flow_mass_in[t],
                to_units=pyunits.pound / pyunits.hr,
            )

    @property
    def default_costing_method(self):
        return self.cost_hydrothermal_gasification

    @staticmethod
    def cost_hydrothermal_gasification(blk):
        """
        General method for costing hydrothermal gasification unit. Capital cost
        is based on the CHG reactor and other wastewater treatment equipment including
        a feed pump, a booster pump, a feed/product exchanger, a fired heater,
        a hydrocyclone, and a product air fin cooler.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        (
            IF_reactor,
            EP_reactor,
            F0_reactor,
            SE_reactor,
            IF_pump,
            EP_pump,
            F0_pump,
            SE_pump,
            IF_booster,
            EP_booster,
            F0_booster,
            SE_booster,
            IF_hydrocyclone,
            EP_hydrocyclone,
            F0_hydrocyclone,
            SE_hydrocyclone,
            IF_cooler,
            EP_cooler,
            F0_cooler,
            SE_cooler,
            IF_exchanger,
            EP_exchanger,
            F0_exchanger,
            Fnew_exchanger,
            SE_exchanger,
            IF_heater,
            EP_heater,
            F0_heater,
            Fnew_heater,
            SE_heater,
        ) = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "installation_factor_reactor",
                "equipment_cost_reactor",
                "base_flowrate_reactor",
                "scaling_exponent_reactor",
                "installation_factor_pump",
                "equipment_cost_pump",
                "base_flowrate_pump",
                "scaling_exponent_pump",
                "installation_factor_booster",
                "equipment_cost_booster",
                "base_flowrate_booster",
                "scaling_exponent_booster",
                "installation_factor_hydrocyclone",
                "equipment_cost_hydrocyclone",
                "base_flowrate_hydrocyclone",
                "scaling_exponent_hydrocyclone",
                "installation_factor_cooler",
                "equipment_cost_cooler",
                "base_flowrate_cooler",
                "scaling_exponent_cooler",
                "installation_factor_exchanger",
                "equipment_cost_exchanger",
                "base_area_exchanger",
                "new_area_exchanger",
                "scaling_exponent_exchanger",
                "installation_factor_heater",
                "equipment_cost_heater",
                "base_heat_duty_heater",
                "new_heat_duty_heater",
                "scaling_exponent_heater",
            ],
        )

        sizing_term_reactor = pyo.units.convert(
            (blk.unit_model.flow_mass_in[t0] / F0_reactor),
            to_units=pyo.units.dimensionless,
        )

        sizing_term_pump = pyo.units.convert(
            (blk.unit_model.flow_mass_in[t0] / F0_pump),
            to_units=pyo.units.dimensionless,
        )

        sizing_term_booster = pyo.units.convert(
            (blk.unit_model.flow_mass_in[t0] / F0_booster),
            to_units=pyo.units.dimensionless,
        )

        sizing_term_hydrocyclone = pyo.units.convert(
            (blk.unit_model.flow_mass_in[t0] / F0_hydrocyclone),
            to_units=pyo.units.dimensionless,
        )

        sizing_term_cooler = pyo.units.convert(
            (blk.unit_model.flow_mass_in[t0] / F0_cooler),
            to_units=pyo.units.dimensionless,
        )

        sizing_term_exchanger = pyo.units.convert(
            (Fnew_exchanger / F0_exchanger),
            to_units=pyo.units.dimensionless,
        )

        sizing_term_heater = pyo.units.convert(
            (Fnew_heater / F0_heater),
            to_units=pyo.units.dimensionless,
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        reactor_cost = pyo.units.convert(
            IF_reactor * EP_reactor * sizing_term_reactor**SE_reactor,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        pump_cost = pyo.units.convert(
            IF_pump * EP_pump * sizing_term_pump**SE_pump,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        booster_cost = pyo.units.convert(
            IF_booster * EP_booster * sizing_term_booster**SE_booster,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        hydrocyclone_cost = pyo.units.convert(
            IF_hydrocyclone
            * EP_hydrocyclone
            * sizing_term_hydrocyclone**SE_hydrocyclone,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        cooler_cost = pyo.units.convert(
            IF_cooler * EP_cooler * sizing_term_cooler**SE_cooler,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        exchanger_cost = pyo.units.convert(
            IF_exchanger * EP_exchanger * sizing_term_exchanger**SE_exchanger,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        heater_cost = pyo.units.convert(
            IF_heater * EP_heater * sizing_term_heater**SE_heater,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        expr = (
            reactor_cost
            + pump_cost
            + booster_cost
            + hydrocyclone_cost
            + cooler_cost
            + exchanger_cost
            + heater_cost
        )

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )
        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.catalyst_flow[t0], "catalyst_HTG"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog


_log = idaeslog.getLogger(__name__)


# when using this file the name "Filtration" is what is imported
@declare_process_block_class("Filtration")
class FiltrationData(UnitModelBlockData):
    """
    Zero order filtration model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Add unit variables
        self.recovery_mass_phase_comp = Var(
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=0.5,
            bounds=(1e-8, 1),
            units=pyunits.dimensionless,
        )
        self.removal_fraction_mass_phase_comp = Var(
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=0.5,
            bounds=(1e-8, 1),
            units=pyunits.dimensionless,
        )

        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
        # Add inlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet
        self.properties_in = self.config.property_package.state_block_class(
            self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        )
        # Add outlet and waste block
        tmp_dict["defined_state"] = False  # outlet and waste block is not an inlet
        self.properties_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )
        self.properties_waste = self.config.property_package.state_block_class(
            self.flowsheet().config.time, doc="Material properties of waste", **tmp_dict
        )

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name="inlet", block=self.properties_in)
        self.add_port(name="outlet", block=self.properties_out)
        self.add_port(name="waste", block=self.properties_waste)

        # Add constraints
        # Usually unit models use a control volume to do the mass, energy, and momentum
        # balances, however, in this example I'm just going to be explicit and write them myself
        @self.Constraint(
            self.config.property_package.component_list,
            doc="Mass balance for components",
        )
        def eq_mass_balance(b, j):
            return (
                b.properties_in[0].flow_mass_phase_comp["Liq", j]
                == b.properties_out[0].flow_mass_phase_comp["Liq", j]
                + b.properties_waste[0].flow_mass_phase_comp["Liq", j]
            )

        @self.Constraint(doc="Isothermal assumption 1")
        def eq_isothermal_1(b):
            return b.properties_in[0].temperature == b.properties_out[0].temperature

        @self.Constraint(doc="Isothermal assumption 2")
        def eq_isothermal_2(b):
            return b.properties_in[0].temperature == b.properties_waste[0].temperature

        @self.Constraint(doc="Isobaric assumption")
        def eq_isobaric(b):
            return b.properties_in[0].pressure == b.properties_out[0].pressure

        # waste pressure should be set by the user

        # Next, add constraints linking variables
        @self.Constraint(
            self.config.property_package.component_list,
            doc="Relationship between removal and recovery",
        )
        def eq_removal_recovery(b, j):
            return (
                b.removal_fraction_mass_phase_comp["Liq", j]
                == 1 - b.recovery_mass_phase_comp["Liq", j]
            )

        @self.Constraint(
            self.config.property_package.component_list,
            doc="Component removal equation",
        )
        def eq_removal_balance(b, j):
            return (
                b.properties_in[0].flow_mass_phase_comp["Liq", j]
                * b.removal_fraction_mass_phase_comp["Liq", j]
                == b.properties_waste[0].flow_mass_phase_comp["Liq", j]
            )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        iscale.set_scaling_factor(self.recovery_mass_phase_comp, 1)
        iscale.set_scaling_factor(self.removal_fraction_mass_phase_comp, 1)

        # transforming constraints
        for j, c in self.eq_mass_balance.items():
            sf = iscale.get_scaling_factor(
                self.properties_in[0].flow_mass_phase_comp["Liq", j]
            )
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_isothermal_1.items():
            sf = iscale.get_scaling_factor(self.properties_in[0].temperature)
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_isothermal_2.items():
            sf = iscale.get_scaling_factor(self.properties_in[0].temperature)
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_isobaric.items():
            sf = iscale.get_scaling_factor(self.properties_in[0].pressure)
            iscale.constraint_scaling_transform(c, sf)

        # other constraints don't need to be transformed
